/*
 * 24.c - 24kHz(640x400 dot)[h CRTC ݒl߂
 * programming 1995/02/14 (c) by Hiramoto Kouji (hiramoto@Angel.nanako.or.jp)
 */

#include <stdio.h>
#include <doslib.h>
#include <iocslib.h>

#define MESSAGE1 "\
             cvX\n\
                 [W]\n\
}CiX[S]  [U]vX\n\
                 [Q]\n\
            c}CiX\n\
\n\
Ȃ - ʕ\̈\n\
         XF1 - ʕ\ʒu\n\
         XF3 - tg|[`\n\
         XF4 - obN|[`\n\
         XF5 - pX\n\
"
#define cls() printf("\033[2J")

#define WIDTH     640               /* ̃hbg */
#define HEIGHT    400               /* c̃hbg */
#define DOTCLOCK8 (0.345 / 1000000) /* 8hbgNbN */
#define DOTCLOCK  (DOTCLOCK8 / 8)   /* hbgNbN */
#define H_STEP    (0.03 / 1000000)  /* ()̕ωP */
#define V_STEP    (0.03 / 1000)     /* c()̕ωP */

#define K_SIZE     0x00
#define K_POSITION 0x01 /* XF1 */
#define K_DUMMY1   0x02 /* XF2 */
#define K_FRONT    0x04 /* XF3 */
#define K_BACK     0x08 /* XF4 */
#define K_PULSE    0x10 /* XF5 */
#define IsPosition (ModifyKey & K_POSITION)
#define IsFront    (ModifyKey & K_FRONT)
#define IsBack     (ModifyKey & K_BACK)
#define IsPulse    (ModifyKey & K_PULSE)

#define DEFAULT_FILENAME "crtc_24"

#define D(a) ((a) ? a : 255)

/* vÕftHgl */
static double H_front = 3.5 / 1000000, H_back = 6.9 / 1000000, H_pulse = 2.4 / 1000000,
              V_front = 0.28 / 1000,   V_back = 1.09 / 1000,   V_pulse = 0.24 / 1000;

static double H_data, H_sync, V_data, V_sync;
static int CRTC[8], ModifyKey;


static void
WriteParam(void)
{
  char s[256];
  FILE *fp;
  int i, j, k;

  printf("Input Filename:");
  fgets(s, 256, stdin);
  if (s[0] == '\n') {
    strcpy(s, DEFAULT_FILENAME);
  } else {
    s[strlen(s) - 1] = '\0';
  }
  if ((fp = fopen(s, "w")) == NULL) {
    printf("%s: ot@CJ܂B\n", s);
    return;
  }
  fprintf(fp, "%f %f %f %f %f %f\n",
	  H_front * 1000000, H_back * 1000000, H_pulse * 1000000,
	  V_front * 1000, V_back * 1000, V_pulse * 1000);
  fprintf(fp, "CRTC= ");
  for (i = 0; i < 8; ++i) {
    j = CRTC[i] / 255;
    k = CRTC[i] % 255;
    fprintf(fp, "%d %d ", D(j), D(k));
  }
  fprintf(fp, "\n");
  fclose(fp);
}

static void
SetCRTC(void)
{
  int i;

  H_data = DOTCLOCK * WIDTH;
  H_sync = H_data + H_front + H_back + H_pulse;
  V_data = H_sync * HEIGHT;
  V_sync = V_data + V_front + V_back + V_pulse;

  CRTC[0] = H_sync / DOTCLOCK8 - 1;
  CRTC[1] = H_pulse / DOTCLOCK8 - 1;
  CRTC[2] = (H_pulse + H_back) / DOTCLOCK8 - 5;
  CRTC[3] = (H_sync - H_front) / DOTCLOCK8 - 5;

  CRTC[4] = V_sync / H_sync - 1;
  CRTC[5] = V_pulse / H_sync - 1;
  CRTC[6] = (V_pulse + V_back) / H_sync - 1;
  CRTC[7] = (V_sync - V_front) / H_sync - 1;

  B_WPOKE((void *)0xe80028,0x0415);
  for (i = 0; i < 8; ++i) {
    B_WPOKE((void *)(0xe80000 + i * 2), CRTC[i]);
  }
  B_WPOKE((void *)0xe82400,0x0004);

  cls();
  printf(MESSAGE1);
  printf("\ng = %f kHz, g = %f Hz\n", 1 / H_sync / 1000, 1 / V_sync);
  printf(": tg|[` %f / obN|[` %f / pX %f\n",
	 H_front * 1000000, H_back * 1000000, H_pulse * 1000000);
  printf(": tg|[` %f / obN|[` %f / pX %f\n",
	 V_front * 1000, V_back * 1000, V_pulse * 1000);
}

static int
DOSgetch(void)
{
  int i;

  while (!(i = K_KEYINP()))
    ;
  ModifyKey = K_SIZE;
  if (K_KEYBIT(0x0a) & 32) {
    ModifyKey |= K_POSITION;
  }
  if (K_KEYBIT(0x0a) & 128) {
    ModifyKey |= K_FRONT;
  }
  if (K_KEYBIT(0x0b) & 1) {
    ModifyKey |= K_BACK;
  }
  if (K_KEYBIT(0x0b) & 2) {
    ModifyKey |= K_PULSE;
  }

  return i;
}

static void
C(double *d, double delta)
{
  double e;

  e = *d + delta;
  if (e < 0) {
    e = 0;
  }
  *d = e;
}

static int
MainLoop(void)
{
  int k;

  k = DOSgetch();

  switch (k) {
  case '2':
    if (IsPosition) {
      C(&V_front, -V_STEP);
      C(&V_back, +V_STEP);
    } else if (IsFront) {
      C(&V_front, -V_STEP);
    } else if (IsBack) {
      C(&V_back, -V_STEP);
    } else if (IsPulse) {
      C(&V_pulse, -V_STEP);
    } else {
      C(&V_front, +V_STEP);
      C(&V_back, +V_STEP);
    }
    break;
  case '8':
    if (IsPosition) {
      C(&V_front, +V_STEP);
      C(&V_back, -V_STEP);
    } else if (IsFront) {
      C(&V_front, +V_STEP);
    } else if (IsBack) {
      C(&V_back, +V_STEP);
    } else if (IsPulse) {
      C(&V_pulse, +V_STEP);
    } else {
      C(&V_front, -V_STEP);
      C(&V_back, -V_STEP);
    }
    break;
  case '4':
    if (IsPosition) {
      C(&H_front, +H_STEP);
      C(&H_back, -H_STEP);
    } else if (IsFront) {
      C(&H_front, -H_STEP);
    } else if (IsBack) {
      C(&H_back, -H_STEP);
    } else if (IsPulse) {
      C(&H_pulse, -H_STEP);
    } else {
      C(&H_front, +H_STEP);
      C(&H_back, +H_STEP);
    }
    break;
  case '6':
    if (IsPosition) {
      C(&H_front, -H_STEP);
      C(&H_back, +H_STEP);
    } else if (IsFront) {
      C(&H_front, +H_STEP);
    } else if (IsBack) {
      C(&H_back, +H_STEP);
    } else if (IsPulse) {
      C(&H_pulse, +H_STEP);
    } else {
      C(&H_front, -H_STEP);
      C(&H_back, -H_STEP);
    }
    break;
  case 'w':
    WriteParam();
    break;
  case 'q':
    return 0;
    break;
  }

  SetCRTC();
  return 1;
}

static void
ReadParam(char *filename)
{
  char s[256];
  FILE *fp;
  float a, b, c, d, e, f;

  if ((fp = fopen(filename, "r")) == NULL) {
    printf("%s: ǂݏot@CJ܂B\n", filename);
    exit(1);
  }

  fgets(s, 256, fp);
  fclose(fp);
  sscanf(s, "%f %f %f %f %f %f", &a, &b, &c, &d, &e, &f);

  H_front = a / 1000000;
  H_back  = b / 1000000;
  H_pulse = c / 1000000;
  V_front = d / 1000;
  V_back  = e / 1000;
  V_pulse = f / 1000;
}

static void
Usage(char *name)
{
	fprintf(stderr, "Usage: %s [-]\n", name);
	exit(1);
}

int
main(int argc, char *argv[])
{
  struct TBOXPTR box;

  switch (argc) {
  case 1:
    /* use Default Parameter */
    break;
  case 2:
    ReadParam(argv[1]);
    break;
  default:
    Usage(argv[0]);
    break;
  }

  C_FNKMOD(2);
  CRTMOD(0x0110);
  B_CONSOL(0,0,80-1,25-1);
  APAGE(0);
  VPAGE(0xf);
  WIPE();
  box.vram_page = 2;
  box.x = 0;
  box.y = 0;
  box.x1 = 639;
  box.y1 = 399;
  box.line_style = 0xffff;
  TXBOX(&box);
  SetCRTC();

  while (MainLoop())
    /* null loop */;

  return 0;
}
